`timescale 1ns / 1ps

`define EOF 32'hFFFF_FFFF
`define NEWLINE 10
`define NULL 0

`ifndef INPUT
`define INPUT "code/house.trace"
`endif
               
`ifndef OUTPUT
`define OUTPUT "code/house.output"
`endif

module testbench_v;

   integer     input_file, output_file, errors, linenum;
   integer     num_cycles, num_branch_stall, num_load_stall, num_divmod_stall;
   

   // Inputs
   reg clk;
   reg rst;
   reg gwe;
   wire [15:0] imem_out;
   wire [15:0] dmem_out;

   // Outputs
   wire [15:0] imem_addr;
   wire [15:0] dmem_addr;
   wire [15:0] dmem_in;
   wire        dmem_we;

   wire [1:0]  test_stall;       // Testbench: is this is stall cycle? (don't compare the test values)
   wire [15:0] test_pc;          // Testbench: program counter
   wire [15:0] test_insn;        // Testbench: instruction bits
   wire        test_regfile_we;  // Testbench: register file write enable
   wire [2:0]  test_regfile_reg; // Testbench: which register to write in the register file 
   wire [15:0] test_regfile_in;  // Testbench: value to write into the register file
   wire        test_nzp_we;      // Testbench: NZP condition codes write enable
   wire [2:0]  test_nzp_in;      // Testbench: value to write to NZP bits
   wire        test_dmem_we;     // Testbench: data memory write enable
   wire [15:0] test_dmem_addr;   // Testbench: address to write memory
   wire [15:0] test_dmem_value;  // Testbench: value to write memory

   reg  [15:0] verify_pc;
   reg  [15:0] verify_insn;
   reg         verify_regfile_we;
   reg  [2:0]  verify_regfile_reg;
   reg  [15:0] verify_regfile_in;
   reg         verify_nzp_we;
   reg  [2:0]  verify_nzp_in;
   reg         verify_dmem_we;
   reg  [15:0] verify_dmem_addr;
   reg  [15:0] verify_dmem_value;

   wire [15:0] vout_dummy;  // video out
   
   // Data and video memory block 
   bram memory (.idclk(clk),
                .iaddr(imem_addr),
                .iout(imem_out),
                .daddr(dmem_addr),
                .dout(dmem_out),
                .din(dmem_in),
                .dwe(dmem_we),
                .vclk(1'b0),
                .vaddr(16'h0000),
                .vout(vout_dummy));
   
   // Instantiate the Unit Under Test (UUT)
   lc4_processor proc (.clk(clk), 
                       .rst(rst),
                       .gwe(gwe),
                       .imem_addr(imem_addr), 
                       .imem_out(imem_out), 
                       .dmem_addr(dmem_addr), 
                       .dmem_in(dmem_in), 
                       .dmem_out(dmem_out), 
                       .dmem_we(dmem_we),
                       .test_stall(test_stall),
                       .test_pc(test_pc),
                       .test_insn(test_insn),
                       .test_regfile_we(test_regfile_we),
                       .test_regfile_reg(test_regfile_reg),
                       .test_regfile_in(test_regfile_in),
                       .test_nzp_we(test_nzp_we),
                       .test_nzp_in(test_nzp_in),
                       .test_dmem_we(test_dmem_we),
                       .test_dmem_addr(test_dmem_addr),
                       .test_dmem_value(test_dmem_value),
                       .switch_data(8'd0)
                       );
   
//   always #5 $display("%d: %d %d %d %h", $time, clk, gwe, rst, test_pc);

   always #5 clk <= ~clk;
   
   /* Because the memory needs a rising edge in order to read, we have
    to slow the processor down to run only on every other cycle. We
    need 2 rising edges to process a load - 1 for the memory, and 1 for
    the regfile. */
//   count #(2) gwe_generator(.clk(clk), .out(gwe));

   always begin 
      #30 gwe <= ~gwe; 
      #10 gwe <= ~gwe;
   end
      

   initial begin
       
      // Initialize Inputs
      gwe = 0;
      clk = 0;
      rst = 1;
      linenum = 0;
      errors = 0;
      num_cycles = 0;
      num_branch_stall = 0;
      num_load_stall = 0;
      num_divmod_stall = 0;
      
      // open the test inputs
      input_file = $fopen(`INPUT, "r");
      if (input_file == `NULL) begin
         $display("Error opening file: %s", `INPUT);
         $finish;
      end

      // open the output file
`ifdef OUTPUT
      output_file = $fopen(`OUTPUT, "w");
      if (output_file == `NULL) begin
         $display("Error opening file: %s", `OUTPUT);
         $finish;
      end
`endif


      #100
      // Wait for global reset to finish
      rst = 0;
      #12;
  
           
      while (10 == $fscanf(input_file, "%h %b %h %h %h %h %h %h %h %h", 
                           verify_pc,
                           verify_insn,
                           verify_regfile_we,
                           verify_regfile_reg,
                           verify_regfile_in,
                           verify_nzp_we,
                           verify_nzp_in,
                           verify_dmem_we,
                           verify_dmem_addr,
                           verify_dmem_value)) begin

         linenum = linenum + 1;

         if (output_file) begin
            $fdisplay(output_file, "%h %b %h %h %h %h %h %h %h %h",
                      verify_pc,
                      verify_insn,
                      verify_regfile_we,
                      verify_regfile_reg,
                      verify_regfile_in,
                      verify_nzp_we,
                      verify_nzp_in,
                      verify_dmem_we,
                      verify_dmem_addr,
                      verify_dmem_value);
         end
         
         while (test_stall !== 2'd0) begin
            num_cycles = num_cycles + 1;
            if (test_stall === 2'd1) begin
               num_branch_stall = num_branch_stall + 1;
            end

            if (test_stall === 2'd2) begin
               num_load_stall = num_load_stall + 1;
            end

            if (test_stall === 2'd3) begin
               num_divmod_stall = num_divmod_stall + 1;
            end
            
            #40;
         end
         num_cycles = num_cycles + 1;
         
         // pc
         if (verify_pc !== test_pc) begin
            $display( "Error at line %d: pc should be %h (but was %h)", 
                      linenum, verify_pc, test_pc);    
            errors = errors + 1;
            $finish;
         end
         
         // insn
         if (verify_insn !== test_insn) begin
            $display( "Error at line %d: insn should be %h (but was %h)", 
                      linenum, verify_insn, test_insn);    
            errors = errors + 1;
            $finish;
         end
         
         // regfile_we
         if (verify_regfile_we !== test_regfile_we) begin
            $display( "Error at line %d: regfile_we should be %h (but was %h)", 
                      linenum, verify_regfile_we, test_regfile_we);    
            errors = errors + 1;
            $finish;
         end
         
         // regfile_reg
         if (verify_regfile_we && verify_regfile_reg !== test_regfile_reg) begin
            $display( "Error at line %d: regfile_reg should be %h (but was %h)", 
                      linenum, verify_regfile_reg, test_regfile_reg);    
            errors = errors + 1;
            $finish;
         end
         
         // regfile_in
         if (verify_regfile_we && verify_regfile_in !== test_regfile_in) begin
            $display( "Error at line %d: regfile_in should be %h (but was %h)", 
                      linenum, verify_regfile_in, test_regfile_in);    
            errors = errors + 1;
            $finish;
         end
         
         // verify_nzp_we
         if (verify_nzp_we !== test_nzp_we) begin
            $display( "Error at line %d: nzp_we should be %h (but was %h)", 
                      linenum, verify_nzp_we, test_nzp_we);    
            errors = errors + 1;
            $finish;
         end
         
         // verify_nzp_in
         if (verify_nzp_we && verify_nzp_in !== test_nzp_in) begin
            $display( "Error at line %d: nzp_in should be %h (but was %h)", 
                      linenum, verify_nzp_in, test_nzp_in);    
            errors = errors + 1;
            $finish;
         end
         
         // verify_dmem_we
         if (verify_dmem_we !== test_dmem_we) begin
            $display( "Error at line %d: dmem_we should be %h (but was %h)", 
                      linenum, verify_dmem_we, test_dmem_we);    
            errors = errors + 1;
            $finish;
         end
         
         // dmem_addr
         if (verify_dmem_addr !== test_dmem_addr) begin
            $display( "Error at line %d: dmem_addr should be %h (but was %h)", 
                      linenum, verify_dmem_addr, test_dmem_addr);    
            errors = errors + 1;
            $finish;
         end
         
         // dmem_value
         if (verify_dmem_value !== test_dmem_value) begin
            $display( "Error at line %d: dmem_value should be %h (but was %h)", 
                      linenum, verify_dmem_value, test_dmem_value);    
            errors = errors + 1;
            $finish;
         end
         
         #40;
      end 
      
      if (input_file) $fclose(input_file); 
      if (output_file) $fclose(output_file);
      $display("Simulation finished: %d test cases %d errors [%s]", linenum, errors, `INPUT);

      if (linenum != num_cycles) begin
         $display("  Instructions:  %d", linenum);
         $display("  Cycles:        %d %d", num_cycles, 1000 * num_cycles / linenum);
         $display("  Branch stalls: %d %d", num_branch_stall, 1000 * num_branch_stall/ linenum);
         $display("  Load stalls:   %d %d", num_load_stall, 1000 * num_load_stall / linenum);
         $display("  ModDiv stalls: %d %d", num_divmod_stall, 1000 * num_divmod_stall / linenum);
      end

      $finish;
   end // initial begin
   
endmodule

